#!/usr/bin/env python3


from enum import Enum
from fp import FP
from FixedPoint import FXfamily, FXnum


class Expression(object):
	def __init__(self, ac, ae, bc, be):
		self._ac = ac
		self._ae = ae
		self._bc = bc
		self._be = be
		self._acFP = FXnum(self._ac, FXfamily(4, 4))
		self._aeFP = FXnum(self._ae, FXfamily(4, 4))
		self._bcFP = FXnum(self._bc, FXfamily(4, 4))
		self._beFP = FXnum(self._be, FXfamily(4, 4))


	def evalFull(self, x):
		p1 = (x ** self._ae) * self._ac
		p2 = (x ** self._be) * self._bc

		return p1 + p2


	def evalFP(self, x):
		xFP = FXnum(x, FXfamily(4, 4))

		p1FP = (xFP ** self._aeFP) * self._acFP
		p2FP = (xFP ** self._beFP) * self._bcFP

		return p1FP + p2FP


	def __eq__(self, other):
		return True if (\
			(self._acFP == other._acFP) and (self._bcFP == other._bcFP) and\
			(self._aeFP == other._aeFP) and (self._beFP == other._beFP)\
		) or (\
			(self._acFP == other._bcFP) and (self._bcFP == other._acFP) and\
			(self._aeFP == other._beFP) and (self._beFP == other._aeFP)\
		) else False


	def __str__(self):
		p1Str = ""
		p2Str = ""
		acFP = self._acFP.scaledval / 16
		aeFP = self._aeFP.scaledval / 16
		bcFP = self._bcFP.scaledval / 16
		beFP = self._beFP.scaledval / 16

		if self._acFP != 0:
			p1Str = "({}x^{})".format(acFP, aeFP)

		if self._bcFP != 0:
			p2Str = "({}x^{})".format(bcFP, beFP)

		if 0 == len(p1Str) and len(p2Str) != 0:
			return p2Str
		elif 0 == len(p2Str) and len(p1Str) != 0:
			return p1Str
		elif 0 == len(p1Str) and 0 == len(p2Str):
			return ""
		else:
			return "{} + {}".format(p1Str, p2Str)


class Expression2Mode(Enum):
	ONE_BY_X = 0
	N_BY_X = 1


class Expression2(object):
	def __init__(self, ac, ae, bc, be):
		self._ac = ac
		self._ae = ae
		self._bc = bc
		self._be = be
		self._acFP = FP(self._ac)
		self._aeFP = FP(self._ae)
		self._bcFP = FP(self._bc)
		self._beFP = FP(self._be)


	def evalFull(self, x):
		p1 = (x ** self._ae) * self._ac
		p2 = (x ** self._be) * self._bc

		return p1 + p2


	def evalFP(self, x, mode=Expression2Mode.ONE_BY_X):
		xFP = FP(x)

		if Expression2Mode.ONE_BY_X == mode:
			p1FP = (xFP ** self._aeFP) * self._acFP
			p2FP = (xFP ** self._beFP) * self._bcFP
		else:
			if self._aeFP._val < 128:
				p1FP = (xFP ** self._aeFP) * self._acFP
			else:
				p1FP = self._acFP / (xFP ** -self._aeFP)

			if self._beFP._val < 128:
				p2FP = (xFP ** self._beFP) * self._bcFP
			else:
				p2FP = self._bcFP / (xFP ** -self._beFP)

		return p1FP + p2FP


	def __eq__(self, other):
		return True if (\
			(self._acFP == other._acFP) and (self._bcFP == other._bcFP) and\
			(self._aeFP == other._aeFP) and (self._beFP == other._beFP)\
		) or (\
			(self._acFP == other._bcFP) and (self._bcFP == other._acFP) and\
			(self._aeFP == other._beFP) and (self._beFP == other._aeFP)\
		) else False


	def __str__(self):
		p1Str = ""
		p2Str = ""

		if self._acFP != 0:
			p1Str = "({}x^{})".format(self._acFP, self._aeFP)

		if self._bcFP != 0:
			p2Str = "({}x^{})".format(self._bcFP, self._beFP)

		if 0 == len(p1Str) and len(p2Str) != 0:
			return p2Str
		elif 0 == len(p2Str) and len(p1Str) != 0:
			return p1Str
		elif 0 == len(p1Str) and 0 == len(p2Str):
			return ""
		else:
			return "{} + {}".format(p1Str, p2Str)


	def binStr(self):
		p1Str = ""
		p2Str = ""

		if self._acFP != 0:
			p1Str = "({}x^{})".format(bin(self._acFP._val & 0b11111111)[2:], bin(self._aeFP._val & 0b11111111)[2:])

		if self._bcFP != 0:
			p2Str = "({}x^{})".format(bin(self._bcFP._val & 0b11111111)[2:], bin(self._beFP._val & 0b11111111)[2:])

		if 0 == len(p1Str) and len(p2Str) != 0:
			return p2Str
		elif 0 == len(p2Str) and len(p1Str) != 0:
			return p1Str
		elif 0 == len(p1Str) and 0 == len(p2Str):
			return ""
		else:
			return "{} + {}".format(p1Str, p2Str)
