#!/usr/bin/env python3


class FP(object):
	_val = 0


	def __init__(self, val):
		shiftedVal = int(float(val) * 16) & 0xff

		if (val > 0 and shiftedVal >= 128) or (val < 0 and shiftedVal < 128):
			raise OverflowError

		self._val = shiftedVal


	def __str__(self):
		return "{}{}".format("-" if self._val >= 128 else "", ((-self)._val / 16.0) if self._val >= 128 else (self._val / 16.0))


	def __neg__(self):
		retVal = FP(0)
		retVal._val = -self._val & 0xff
		return retVal


	def __add__(self, other):
		selfSign = -1 if self._val >= 128 else 1
		otherSign = -1 if other._val >= 128 else 1
		shiftedVal = (self._val + other._val) & 0xff
		shiftedSign = -1 if shiftedVal >= 128 else 1

		if (selfSign > 0 and otherSign > 0 and shiftedSign < 0) or (selfSign < 0 and otherSign < 0 and shiftedSign > 0):
			raise OverflowError

		retVal = FP(0)
		retVal._val = shiftedVal
		return retVal


	def __sub__(self, other):
		return self + (-other)


	def __mul__(self, other):
		selfSign = -1 if self._val >= 128 else 1
		otherSign = -1 if other._val >= 128 else 1
		posSelfVal = self._val if selfSign > 0 else (-self._val & 0xff)
		posOtherVal = other._val if otherSign > 0 else (-other._val & 0xff)
		shiftedVal = ((posSelfVal * posOtherVal) >> 4) & 0xffff

		if shiftedVal >= 256:
			raise OverflowError

		shiftedVal = shiftedVal & 0xff

		return FP(shiftedVal / 16.0) if selfSign == otherSign else FP(-(shiftedVal / 16.0))


	def __truediv__(self, other):
		selfSign = -1 if self._val >= 128 else 1
		otherSign = -1 if other._val >= 128 else 1
		posSelfVal = self._val if selfSign > 0 else (-self._val & 0xff)
		posOtherVal = other._val if otherSign > 0 else (-other._val & 0xff)
		shiftedVal = int((posSelfVal << 4) / posOtherVal) & 0xffff

		if shiftedVal >= 256:
			raise OverflowError

		shiftedVal = shiftedVal & 0xff

		return FP(shiftedVal / 16.0) if selfSign == otherSign else FP(-(shiftedVal / 16.0))


	__div__ = __truediv__


	def __pow__(self, other, modulus=None):
		if modulus is not None:
			raise NotImplementedError

		otherSign = -1 if other._val >= 128 else 1
		posOtherVal = other._val if otherSign > 0 else (-other._val & 0xff)

		if (posOtherVal & 0xf) != 0:
			raise NotImplementedError

		posOtherVal = posOtherVal >> 4
		accVal = FP(1)

		#if 0 == other._val:
		#	return accVal
		#elif otherSign < 0:
		#	for i in range(0, posOtherVal):
		#		accVal = accVal / self
		#else:
		#	for i in range(0, posOtherVal):
		#		accVal = accVal * self
		if 0 == other._val:
			return accVal
		else:
			for i in range(0, posOtherVal):
				accVal = accVal * self

			if otherSign < 0:
				accVal = FP(1) / accVal

		return accVal
